import gradio as gr
from megabots import create_interface


def test_create_interface():
    # create a mock Bot object
    class MockBot:
        def ask(self, question: str):
            return "Answer"

    # create a mock example
    example = [["What is your name?"], ["My name is Bot."]]

    # call the function with the mock bot and example
    interface = create_interface(MockBot(), examples=example)

    # check if the interface has the correct properties
    assert isinstance(interface, gr.Blocks)



from pytest import raises
from megabots import memory
from megabots.memory import ConversationBuffer, ConversationBufferWindow


def test_memory_conversation_buffer():
    mem = memory(name="conversation-buffer")
    assert isinstance(mem, ConversationBuffer)


def test_memory_conversation_buffer_window():
    mem = memory(name="conversation-buffer-window", k=10)
    assert isinstance(mem, ConversationBufferWindow)


def test_memory_unsupported_name():
    with raises(ValueError, match=r"Memory invalid-name is not supported."):
        memory(name="invalid-name")


def test_memory_no_name():
    with raises(
        RuntimeError, match=r"Impossible to instantiate memory without a name."
    ):
        memory(name=None)



import gradio as gr
from fastapi import FastAPI
from megabots.bot import Bot


def create_api(bot: Bot):
    app = FastAPI()

    @app.get("/v1/ask/{question}")
    async def ask(question: str):
        answer = bot.ask(question)
        return {"answer": answer}

    return app


def create_interface(bot_instance: Bot, examples: list[list[str]] = []):
    with gr.Blocks() as interface:
        chatbot = gr.Chatbot([], elem_id="chatbot").style(height=750)
        msg = gr.Textbox(
            show_label=False,
            placeholder="Enter text and press enter, or upload an image",
        ).style(container=False)
        clear = gr.Button("Clear")

        def user(user_message, history):
            return "", history + [[user_message, None]]

        def bot(history):
            response = bot_instance.ask(history[-1][0])
            history[-1][1] = response
            return history

        msg.submit(user, [msg, chatbot], [msg, chatbot], queue=False).then(
            bot, chatbot, chatbot
        )
        clear.click(lambda: None, None, chatbot, queue=False)

    return interface



from typing import Type, TypeVar
from langchain.vectorstores import Milvus
from abc import ABC


class MilvusVectorStore:
    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port
        self.client = Milvus


class ChromaVectorStore:
    pass


# Generic type variable for all vectorstores
VectorStore = type("VectorStore", (MilvusVectorStore, ChromaVectorStore), {})


SUPPORTED_VECTORSTORES = {
    "milvus": {
        "impl": MilvusVectorStore,
        "default": {"host": "localhost", "port": 19530},
    }
}


def vectorstore(
    name: str, host: str | None = None, port: int | None = None
) -> VectorStore:
    """Return a vectorstore object."""

    if name is None:
        raise RuntimeError("Impossible to instantiate a vectorstore without a name.")

    if name not in SUPPORTED_VECTORSTORES:
        raise ValueError(f"Vectorstore {name} is not supported.")

    return SUPPORTED_VECTORSTORES[name]["impl"](
        host=host or SUPPORTED_VECTORSTORES[name]["default"]["host"],
        port=port or SUPPORTED_VECTORSTORES[name]["default"]["port"],
    )



import json
from fastapi.testclient import TestClient
from megabots import bot, create_api

qnabot = bot("qna-over-docs", index="./examples/files")
app = create_api(qnabot)

client = TestClient(app)


def test_successful_response():
    response = client.get("/v1/ask/What is your name?")
    assert response.status_code == 200
    assert "answer" in response.json()
    assert isinstance(response.json()["answer"], str)


def test_missing_question_parameter():
    response = client.get("/v1/ask/")
    assert response.status_code == 404
    assert response.json() == {"detail": "Not Found"}



from megabots.vectorstore import VectorStore, vectorstore
from megabots.memory import Memory, memory
from megabots.bot import Bot, bot
from megabots.prompt import prompt
from megabots.utils import create_api, create_interface


from dotenv import load_dotenv

load_dotenv()



from setuptools import setup, find_packages

VERSION = "0.0.11"

setup(
    name="megabots",
    version=VERSION,
    packages=find_packages(),
    install_requires=[
        "langchain",
        "tiktoken",
        "unstructured",
        "fastapi",
        "faiss-cpu",
        "pdfminer.six",
    ],
    author="Megaklis Vasilakis",
    author_email="megaklis.vasilakis@gmail.com",
    description="🤖 Megabots provides State-of-the-art, production ready bots made mega-easy, so you don't have to build them from scratch 🤯 Create a bot, now 🫵",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/momegas/megabots",
    license="MIT",
    classifiers=[
        # Choose appropriate classifiers from
        # https://pypi.org/classifiers/
        "Development Status :: 4 - Beta"
    ],
)



from typing import Any
from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI
from langchain.embeddings import OpenAIEmbeddings
from langchain.chains.qa_with_sources import load_qa_with_sources_chain
from langchain.vectorstores.faiss import FAISS
import pickle
import os
from langchain.prompts import PromptTemplate
from langchain.chains.question_answering import load_qa_chain
from langchain.chains.conversational_retrieval.prompts import QA_PROMPT
from langchain.document_loaders import DirectoryLoader
from megabots.prompt import QA_MEMORY_PROMPT
from megabots.vectorstore import VectorStore
from megabots.memory import Memory
import megabots


class Bot:
    def __init__(
        self,
        model: str | None = None,
        prompt: PromptTemplate | None = None,
        index: str | None = None,
        sources: bool | None = False,
        vectorstore: VectorStore | None = None,
        memory: Memory | None = None,
        verbose: bool = False,
        temperature: int = 0,
    ):
        self.vectorstore = vectorstore
        self.memory = memory
        self.prompt = prompt or QA_MEMORY_PROMPT if self.memory else QA_PROMPT
        self.select_model(model, temperature)
        self.create_loader(index)
        self.load_or_create_index(index, vectorstore)

        # Load the question-answering chain for the selected model
        self.chain = self.create_chain(sources=sources, verbose=verbose)

    def create_chain(
        self,
        sources: bool | None = False,
        verbose: bool = False,
    ):
        # TODO: Changing the prompt here is not working. Leave it as is for now.
        # Reference: https://github.com/hwchase17/langchain/issues/2858
        if sources:
            return load_qa_with_sources_chain(
                self.llm,
                chain_type="stuff",
                memory=self.memory.memory if self.memory else None,
                verbose=verbose,
            )
        return load_qa_chain(
            self.llm,
            chain_type="stuff",
            verbose=verbose,
            prompt=self.prompt,
            memory=self.memory.memory if self.memory else None,
        )

    def select_model(self, model: str | None, temperature: float):
        # Select and set the appropriate model based on the provided input
        if model is None or model == "gpt-3.5-turbo":
            print("Using model: gpt-3.5-turbo")
            self.llm = ChatOpenAI(temperature=temperature)

        if model == "text-davinci-003":
            print("Using model: text-davinci-003")
            self.llm = OpenAI(temperature=temperature)

    def create_loader(self, index: str | None):
        # Create a loader based on the provided directory (either local or S3)
        if index is None:
            raise RuntimeError(
                """
            Impossible to find a valid index. 
            Either provide a valid path to a pickle file or a directory.               
            """
            )
        self.loader = DirectoryLoader(index, recursive=True)

    def load_or_create_index(self, index: str, vectorstore: VectorStore | None = None):
        # Load an existing index from disk or create a new one if not available
        if vectorstore is not None:
            self.search_index = vectorstore.client.from_documents(
                self.loader.load_and_split(),
                OpenAIEmbeddings(),
                connection_args={"host": vectorstore.host, "port": vectorstore.port},
            )
            return

        # Is pickle
        if index is not None and "pkl" in index or "pickle" in index:
            print("Loading path from pickle file: ", index, "...")
            with open(index, "rb") as f:
                self.search_index = pickle.load(f)
            return

        # Is directory
        if index is not None and os.path.isdir(index):
            print("Creating index...")
            self.search_index = FAISS.from_documents(
                self.loader.load_and_split(), OpenAIEmbeddings()
            )
            return

        raise RuntimeError(
            """
            Impossible to find a valid index. 
            Either provide a valid path to a pickle file or a directory.               
            """
        )

    def save_index(self, index_path: str):
        # Save the index to the specified path
        with open(index_path, "wb") as f:
            pickle.dump(self.search_index, f)

    def ask(self, question: str, k=1) -> str:
        # Retrieve the answer to the given question and return it
        input_documents = self.search_index.similarity_search(question, k=k)
        answer = self.chain.run(input_documents=input_documents, question=question)
        return answer


SUPPORTED_TASKS = {
    "qna-over-docs": {
        "impl": Bot,
        "default": {
            "model": "gpt-3.5-turbo",
            "temperature": 0,
            "index": "./index",
            "input_variables": ["context", "question"],
        },
    }
}

SUPPORTED_MODELS = {}


def bot(
    task: str | None = None,
    *,
    model: str | None = None,
    index: str | None = None,
    prompt: str | None = None,
    memory: str | Memory | None = None,
    vectorstore: str | VectorStore | None = None,
    verbose: bool = False,
    temperature: int = 0,
) -> Bot:
    """Instanciate a bot based on the provided task. Each supported tasks has it's own default sane defaults.

    Args:
        task (str | None, optional): The given task. Can be one of the SUPPORTED_TASKS.

        model (str | None, optional): Model to be used. Can be one of the SUPPORTED_MODELS.

        index (str | None, optional): Data that the model will load and store index info.
        Can be either a local file path, a pickle file, or a url of a vector database.
        By default it will look for a local directory called "files" in the current working directory.

        prompt (str | None, optional): The prompt that the bot will take in. Mark variables like this: {variable}.
        Variables are context, question, and history if the bot has memory.

        vectorstore: (str | VectorStore | None, optional): The vectorstore that the bot will save the index to.
        If only a string is passed, the defaults values willl be used.

        verbose (bool, optional): Verbocity. Defaults to False.

        temperature (int, optional): Temperature. Defaults to 0.

    Raises:
        RuntimeError: _description_
        ValueError: _description_

    Returns:
        Bot: Bot instance
    """

    if task is None:
        raise RuntimeError("Impossible to instantiate a bot without a task.")
    if task not in SUPPORTED_TASKS:
        raise ValueError(f"Task {task} is not supported.")

    task_defaults = SUPPORTED_TASKS[task]["default"]

    if memory is not None:
        task_defaults["input_variables"].append("history")

    return SUPPORTED_TASKS[task]["impl"](
        model=model or task_defaults["model"],
        index=index or task_defaults["index"],
        prompt=None
        if prompt is None
        else PromptTemplate(
            template=prompt, input_variables=task_defaults["input_variables"]
        ),
        temperature=temperature,
        verbose=verbose,
        vectorstore=megabots.vectorstore(vectorstore)
        if isinstance(vectorstore, str)
        else vectorstore,
        memory=megabots.memory(memory) if isinstance(memory, str) else memory,
    )



from langchain.memory import ConversationBufferMemory, ConversationBufferWindowMemory


class ConversationBuffer:
    def __init__(self):
        self.memory = ConversationBufferMemory(input_key="question")


class ConversationBufferWindow:
    def __init__(self, k: int):
        self.k: int = k
        self.memory = ConversationBufferWindowMemory(k=self.k, input_key="question")


SUPPORTED_MEMORY = {
    "conversation-buffer": {
        "impl": ConversationBuffer,
        "default": {},
    },
    "conversation-buffer-window": {
        "impl": ConversationBufferWindow,
        "default": {"k": 3},
    },
}


Memory = type("Memory", (ConversationBuffer, ConversationBufferWindow), {})


def memory(
    name: str = "conversation-buffer-window",
    k: int | None = None,
) -> Memory:
    if name is None:
        raise RuntimeError("Impossible to instantiate memory without a name.")

    if name not in SUPPORTED_MEMORY:
        raise ValueError(f"Memory {name} is not supported.")

    cl = SUPPORTED_MEMORY[name]["impl"]

    if name == "conversation-buffer-window":
        return cl(k=k)

    return SUPPORTED_MEMORY[name]["impl"]()
